#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides a processor that reads a GRO file.

See also
--------
:mod:`vermouth.gmx.gro`
"""


from ..gmx import gro
from .processor import Processor

class GROInput(Processor):
    def __init__(self, filename, exclude=(), ignh=False):
        super().__init__()
        self.filename = filename
        self.exclude = exclude
        self.ignh = ignh

    def run_system(self, system):
        molecule = gro.read_gro(self.filename,
                                exclude=self.exclude,
                                ignh=self.ignh)
        system.add_molecule(molecule)
